/*
 *	sortnews.c
 *
 *	sort .newsrc
 */

#include <stdio.h>
#include <pwd.h>
#include <string.h>
#include <sys/types.h>
#include "sortnews.h"
#include "term.h"
#include "version.h"

static VOID usage __P_((char *));
static VOID error __P_((int));
static int cdhome __P_((VOID_A));
static char *fgetline __P_((FILE *));
static int fcopyline __P_((long, int, FILE *, FILE *));
static ngtree *newtree __P_((char *, long, int, ngtree *));
static ngtree *addtree __P_((ngtree *, char *, long, int));
static VOID inserttree __P_((ngtree *, long, ngtree *));
static ngtree *duplparent __P_((ngtree *));
static int cmptree __P_((CONST VOID_P, CONST VOID_P));
static int cmpparent __P_((ngtree *, ngtree *));
static int addng __P_((char *, long));
static VOID freetree __P_((ngtree *));
static VOID catpath __P_((ngtree *));
static ngtree *concatng __P_((ngtree *, ngtree *));
static ngtree *combineng __P_((ngtree *, ngtree *));
static ngtree *divideng __P_((ngtree *, long));
static VOID *deleteng __P_((ngtree *, long));
static ngtree **pushng __P_((ngtree *, long, ngtree **, long));
static ngtree *popng __P_((ngtree *, long, ngtree *));
static VOID restorestack __P_((ngtree **, long));
static int backng __P_((ngtree *, ngtree *, ngtree **, long));
static VOID sortng __P_((ngtree *));
static VOID r_sortng __P_((ngtree *));
static VOID mergeng __P_((ngtree *));
static VOID r_mergeng __P_((ngtree *));
static VOID r_setflagng __P_((ngtree *, int));
static int readrc __P_((VOID_A));
static VOID title __P_((VOID_A));
static int listupgroup __P_((ngtree *, long));
static VOID listupstack __P_((ngtree **, long));
static int helpline __P_((int, char *, char *));
static VOID helpwin __P_((VOID_A));
static int _dumpng __P_((ngtree *, FILE *, FILE *));
static VOID dumpng __P_((VOID_A));
static int getyesno __P_((char *));
static VOID browsegroup __P_((VOID_A));
static int evaloption __P_((int, char *[]));

static ngtree *topng = NULL;
static long *options = NULL;
static char *newsrcfile = NULL;
static char *backupfile = NULL;
static char *temporaryfile = NULL;
static int flags = 0;
static int isearch = 0;


static VOID usage(name)
char *name;
{
	fputs("usage: ", stderr);
	fputs(name, stderr);
	fputs(" [-imbh] [newsrc_file]\n", stderr);
	fputs("    -i: interactive edit mode\n", stderr);
	fputs("    -m: not sort but merge\n", stderr);
	fputs("    -u: not save backup file (unsafe)\n", stderr);
	fputs("    -h: show this\n", stderr);
	exit(1);
}

static VOID error(n)
int n;
{
	endterm();
	inittty(1);
	fputc('\007', stderr);
	switch (n) {
		case ER_NOMEM:
			fputs("Not enough memory.\n", stderr);
			break;
		case ER_NOREAD:
			fputs(newsrcfile, stderr);
			fputs(": cannot read.\n", stderr);
			break;
		case ER_NOWRITE:
			fputs(newsrcfile, stderr);
			fputs(": cannot write.\n", stderr);
			break;
		default:
			break;
	}
	exit(1);
}

static int cdhome(VOID_A)
{
	struct passwd *pwd;
	char *cp;

	if (!(cp = (char *)getenv("HOME"))) {
		if (!(pwd = getpwuid(getuid()))) return(-1);
		cp = pwd -> pw_dir;
	}
	if (chdir(cp)) return(-1);
	return(0);
}

static char *fgetline(fp)
FILE *fp;
{
	char *cp;
	long i, size;
	int c;

	cp = (char *)malloc(size = BUFUNIT);

	for (i = 0; (c = fgetc(fp)) != '\n'; i++) {
		if (c == EOF) {
			if (ferror(fp)) error(ER_NOREAD);
			if (!i) {
				free(cp);
				return(NULL);
			}
			break;
		}
		if (i + 1 >= size) {
			if (size < (1 << (sizeof(long) * 8 - 2)))
				cp = (char *)realloc(cp, (size *= 2));
			else {
				free(cp);
				cp = NULL;
			}
			if (!cp) error(ER_NOMEM);
		}
		cp[i] = c;
	}
	cp[i] = '\0';
	return(cp);
}

static int fcopyline(offset, flag, fpin, fpout)
long offset;
int flag;
FILE *fpin, *fpout;
{
	int c;

	if (fseek(fpin, offset, 0) < 0) return(-1);

	c = fgetc(fpin);
	while (c != EOF) {
		if (flag && (c == ':' || c == '!')) {
			c = flag;
			flag = '\0';
		}
		if (fputc(c, fpout) == EOF) return(-2);
		if (c != '\n') c = fgetc(fpin);
		else {
			if ((c = fgetc(fpin)) == EOF) return(0);
			if (c != ' ' && c != '\t') break;
		}
	}
	if (c == EOF && ferror(fpin)) return(-1);
	return(0);
}

static ngtree *newtree(name, len, flag, parent)
char *name;
long len;
int flag;
ngtree *parent;
{
	ngtree *ng;
	char *cp;

	ng = (ngtree *)malloc(sizeof(ngtree));
	cp = (char *)malloc(len + 2);
	if (!ng || !cp) error(ER_NOMEM);

	*cp = flag;
	memcpy(cp + 1, name, len);
	cp[len + 1] = '\0';
	ng -> name = cp;
	ng -> offset = -1;
	ng -> max = 0;
	if (!parent) ng -> sub = NULL;
	else {
		ng -> sub = (ngtree **)malloc(sizeof(ngtree *));
		if (!(ng -> sub)) error(ER_NOMEM);
		ng -> sub[0] = parent;
	}
	return(ng);
}

static ngtree *addtree(ng, name, len, flag)
ngtree *ng;
char *name;
long len;
int flag;
{
	if (!(ng -> sub))
		ng -> sub = (ngtree **)malloc(BUFUNIT * sizeof(ngtree *));
	else if (!((ng -> max) % BUFUNIT))
		ng -> sub = (ngtree **)realloc(ng -> sub,
			b_size(ng -> max, ngtree *));
	if (!(ng -> sub)) error(ER_NOMEM);

	ng -> sub[ng -> max] =
		(len >= 0) ? newtree(name, len, flag, ng) : (ngtree *)name;
	return(ng -> sub[(ng -> max)++]);
}

static VOID inserttree(ng, no, inserted)
ngtree *ng;
long no;
ngtree *inserted;
{
	long i;

	addtree(ng, NULL, -1, '\0');
	for (i = ng -> max - 1; i > no; i--)
		ng -> sub[i] = ng -> sub[i - 1];
	ng -> sub[i] = inserted;
}

static ngtree *duplparent(ng)
ngtree *ng;
{
	ngtree *parent, *tmp;

	parent = ng -> sub[0];
	if (!(parent -> name)) return(NULL);
	tmp = newtree(ngname(parent), strlen(ngname(parent)),
		ngflag(parent), NULL);
	addtree(tmp, (char *)(parent -> sub[0]), -1, '\0');
	if (*ngname(ng) == '.') tmp -> max = 0;
	else addtree(tmp, (char *)ng, -1, '\0');
	return(tmp);
}

static int cmptree(vp1, vp2)
CONST VOID_P vp1;
CONST VOID_P vp2;
{
	return(ngcmp(*(ngtree **)vp1, *(ngtree **)vp2));
}

static int cmpparent(ng1, ng2)
ngtree *ng1, *ng2;
{
	int i;

	if (ng1 == ng2) return(0);
	else if (!(ng1 -> name)) return(-1);
	else if (!(ng2 -> name)) return(1);
	else if ((i = ngcmp(ng1, ng2))) return(i);
	else return(cmpparent(ng1 -> sub[0], ng2 -> sub[0]));
}

static int addng(str, offset)
char *str;
long offset;
{
	ngtree *ng, *parent;
	char *cp;
	long i, n;

	ng = topng;
	parent = NULL;
	for (cp = str; *cp; cp++) {
		if (*cp != '.' && *cp != ':' && *cp != '!') continue;
		n = -1;
		for (i = startng(ng); i < ng -> max; i++) {
			if (!strncmp(ngname(ng -> sub[i]), str, cp - str)
			&& !(ngname(ng -> sub[i])[cp - str])) n = i;
		}

		if (n >= 0 && n == ng -> max - 1) {
			parent = ng;
			ng = ng -> sub[n];
			if (*cp != '.') {
				addtree(ng, ".", 1, *cp);
				ng -> offset = offset;
				break;
			}
		}
		else {
			if (!(ng -> max)) {
				if (parent)
					addtree(ng, (char *)parent, -1, '\0');
				if (ng -> offset >= 0)
					addtree(ng, ".", 1, ngflag(ng));
			}
			parent = ng;
			ng = addtree(ng, str, cp - str, *cp);
			if (*cp != '.') {
				ng -> offset = offset;
				break;
			}
		}
		str = cp + 1;
	}
	if (ng == topng) return(-1);
	return(0);
}

static VOID freetree(ng)
ngtree *ng;
{
	long i;

	if (ng -> sub) {
		for (i = startng(ng); i < ng -> max; i++) {
			if ((ng -> sub[i]) -> max) freetree(ng -> sub[i]);
			else {
				free((ng -> sub[i]) -> name);
				free((ng -> sub[i]) -> sub);
				free(ng -> sub[i]);
			}
		}
		free(ng -> sub);
	}
	if (ng -> name) free(ng -> name);
	free(ng);
}

static VOID catpath(ng)
ngtree *ng;
{
	if (!ng || ng == topng) cputs2("(top)");
	else {
		if (ng -> sub[0] != topng) {
			catpath(ng -> sub[0]);
			if (*ngname(ng) == '.') return;
			putch2('.');
		}
		cputs2(ngname(ng));
	}
}

static ngtree *concatng(ng1, ng2)
ngtree *ng1, *ng2;
{
	ngtree **tmp;
	long i, max;

	if (!(ng2 -> max)) {
		addtree(ng1, ".", 1, ngflag(ng2));
		ng1 -> offset = ng2 -> offset;
	}
	else if (!(ng1 -> max)) {
		tmp = ng1 -> sub;
		inserttree(ng2, 1, newtree(".", 1, ngflag(ng1), ng1));
		for (i = 1; i < ng2 -> max; i++)
			(ng2 -> sub[i]) -> sub[0] = ng1;
		ng1 -> sub = ng2 -> sub;
		ng1 -> max = ng2 -> max;
		ng2 -> sub = tmp;
	}
	else {
		if (!ngcmp(ng1 -> sub[ng1 -> max - 1], ng2 -> sub[1])) {
			ng2 -> sub[1] = concatng(ng1 -> sub[--(ng1 -> max)],
				ng2 -> sub[1]);
		}
		max = ng1 -> max + ng2 -> max - 1;
		ng1 -> sub = (ngtree **)realloc(ng1 -> sub,
			b_size(max, ngtree *));
		if (!ng1 -> sub) error(ER_NOMEM);

		for (i = 1; i < ng2 -> max; i++) {
			(ng2 -> sub[i]) -> sub[0] = ng1;
			ng1 -> sub[i + ng1 -> max - 1] = ng2 -> sub[i];
		}
		ng1 -> max = max;

		if (ng2 -> offset >= 0) ng1 -> offset = ng2 -> offset;
	}
	if (ng2 -> sub) free(ng2 -> sub);
	if (ng2 -> name) free(ng2 -> name);
	free(ng2);
	return(ng1);
}

static ngtree *combineng(ng1, ng2)
ngtree *ng1, *ng2;
{
	ngtree **tmp;
	long i, j, n;
	int c;

	if (!(ng2 -> max)) {
		inserttree(ng1, 1, newtree(".", 1, ngflag(ng2), ng1));
		ng1 -> offset = ng2 -> offset;
	}
	else if (!(ng1 -> max)) {
		tmp = ng1 -> sub;
		inserttree(ng2, 1, newtree(".", 1, ngflag(ng1), ng1));
		for (i = 1; i < ng2 -> max; i++)
			(ng2 -> sub[i]) -> sub[0] = ng1;
		ng1 -> sub = ng2 -> sub;
		ng1 -> max = ng2 -> max;
		ng2 -> sub = tmp;
	}
	else {
		for (i = 1; i < ng2 -> max; i++) {
			n = 0;
			for (j = 1; j < ng1 -> max; j++) {
				if (!(c = ngcmp(ng2 -> sub[i], ng1 -> sub[j])))
					break;
				if (c > 0) n = j;
			}
			if (j < ng1 -> max)
				combineng(ng1 -> sub[j], ng2 -> sub[i]);
			else {
				(ng2 -> sub[i]) -> sub[0] = ng1;
				inserttree(ng1, n + 1, ng2 -> sub[i]);
			}
		}
		if (ng2 -> offset >= 0) ng1 -> offset = ng2 -> offset;
	}
	if (ng2 -> sub) free(ng2 -> sub);
	if (ng2 -> name) free(ng2 -> name);
	free(ng2);
	return(ng1);
}

static ngtree *divideng(ng, no)
ngtree *ng;
long no;
{
	ngtree *tmp;
	long i, max;

	max = ng -> max - no + 1;
	tmp = newtree(ngname(ng), strlen(ngname(ng)), ngflag(ng), NULL);
	if (ng -> offset) {
		for (i = no; i < ng -> max; i++)
		if (*ngname(ng -> sub[i]) == '.') {
			tmp -> offset = ng -> offset;
			ng -> offset = -1;
			break;
		}
	}
	tmp -> sub = (ngtree **)malloc(b_size(max, ngtree **));
	tmp -> sub[0] = ng -> sub[0];
	for (i = 1; i < max; i++) {
		tmp -> sub[i] = ng -> sub[no + i - 1];
		(tmp -> sub[i]) -> sub[0] = tmp;
	}
	tmp -> max = max;
	ng -> max = no;
	return(tmp);
}

static VOID *deleteng(ng, no)
ngtree *ng;
long no;
{
	long i;

	i = 1;
	if (no > startng(ng) && no < ng -> max -1
	&& !ngcmp(ng -> sub[no - 1], ng -> sub[no + 1])) {
		concatng(ng -> sub[no - 1], ng -> sub[no + 1]);
		(ng -> max)--;
		i = 2;
	}
	(ng -> max)--;
	for (; no < ng -> max; no++) ng -> sub[no] = ng -> sub[no + i];
}

static ngtree **pushng(ng, no, stack, sp)
ngtree *ng;
long no;
ngtree **stack;
long sp;
{
	long i;

	if (!stack) stack = (ngtree **)malloc(BUFUNIT * sizeof(ngtree *));
	else if (!(sp % BUFUNIT))
		stack = (ngtree **)realloc(stack, b_size(sp, ngtree *));
	if (!stack) error(ER_NOMEM);

	stack[sp] = ng -> sub[no];

	deleteng(ng, no);
	return(stack);
}

static ngtree *popng(ng, no, poped)
ngtree *ng;
long no;
ngtree *poped;
{
	ngtree *tr, *tmp;
	long i;

	for (tr = poped; tr; tr = tr -> sub[1]) {
		if (!cmpparent(ng, tr -> sub[0])) {
			while (poped != tr) {
				tmp = poped -> sub[1];
				if (poped -> name) free(poped -> name);
				if (poped -> sub) free(poped -> sub);
				free(poped);
				poped = tmp;
			}
			if (no > startng(ng)
			&& !ngcmp(tr, ng -> sub[no - 1]))
				concatng(ng -> sub[no - 1], tr);
			else if (no < ng -> max
			&& !ngcmp(tr, ng -> sub[no]))
				ng -> sub[no] = concatng(tr, ng -> sub[no]);
			else {
				tr -> sub[0] = ng;
				inserttree(ng, no, tr);
			}
			return(ng);
		}
		if (tr -> max != 2) break;
	}

	if (*ngname(poped) == '.') {
		tmp = poped;
		if ((tmp -> sub[0]) -> max == 2
		&& tmp == (tmp -> sub[0]) -> sub[1]) {
			poped = poped -> sub[0];
			poped -> max = 0;
		}
		else {
			poped = duplparent(poped);
			poped -> offset = (tmp -> sub[0]) -> offset;
			(tmp -> sub[0]) -> offset = -1;
			ngflag(poped) = ngflag(tmp);
		}
		free(tmp -> name);
		free(tmp -> sub);
		free(tmp);
	}
	if ((poped -> sub[0]) -> name) {
		if ((poped -> sub[0]) -> max != 2
		|| poped != (poped -> sub[0]) -> sub[1])
			poped -> sub[0] = duplparent(poped);
		return(popng(ng, no, poped -> sub[0]));
	}

	for (i = startng(ng -> sub[0]); i < (ng -> sub[0]) -> max; i++)
		if (ng == (ng -> sub[0]) -> sub[i]) break;
	if (no <= startng(ng)) return(popng(ng -> sub[0], i, poped));
	if (no >= ng -> max) return(popng(ng -> sub[0], i + 1, poped));

	inserttree(ng -> sub[0], i + 1, divideng(ng, no));
	ng = ng -> sub[0];
	tmp = popng(ng, i + 1, poped);
	if (ng -> max > i + 1 && !ngcmp(ng -> sub[i], ng -> sub[i + 1])) {
		concatng(ng -> sub[i], ng -> sub[i + 1]);
		(ng -> max)--;
		for (i++ ; i < ng -> max; i++) ng -> sub[i] = ng -> sub[i + 1];
	}
	return(tmp);
}

static VOID restorestack(stack, sp)
ngtree **stack;
long sp;
{
	ngtree *ng;
	long i;

	for (i = sp - 1; i >= 0; i--) {
		for (ng = stack[i]; ng -> name; ng = ng -> sub[0])
			if ((ng -> sub[0]) -> max != 2
			|| ng != (ng -> sub[0]) -> sub[1]) break;
		if (!((ng -> sub[0]) -> max)) {
			(ng -> sub[0]) -> max = 1;
			addtree(ng -> sub[0], ".", 1, ngflag(ng -> sub[0]));
		}
		addtree(ng -> sub[0], (char *)ng, -1, '\0');
	}
}

static int backng(ng, back, stack, sp)
ngtree *ng, *back, **stack;
long sp;
{
	ngtree *tmp;
	long i, j;

	i = startng(back);
	while (ng -> name && ng != back) {
		tmp = ng -> sub[0];
		for (i = startng(tmp); i < tmp -> max; i++)
			if (tmp -> sub[i] == ng) break;
		ng = tmp;
		if (i >= ng -> max) continue;

		if ((ng -> sub[i]) -> max > 2) continue;
		if ((ng -> sub[i]) -> max == 2) {
			if (*ngname((ng -> sub[i]) -> sub[1]) != '.')
				continue;
			if (((ng -> sub[i]) -> sub[1]) -> name)
				free(((ng -> sub[i]) -> sub[1]) -> name);
			if (((ng -> sub[i]) -> sub[1]) -> sub)
				free(((ng -> sub[i]) -> sub[1]) -> sub);
			free((ng -> sub[i]) -> sub[1]);
			(ng -> sub[i]) -> max = 0;
			continue;
		}

		for (j = sp - 1; j >= 0; j--) {
			for (tmp = stack[j]; tmp -> name; tmp = tmp -> sub[0])
				if (ng -> sub[i] == tmp -> sub[0]) break;
			if (tmp -> name) break;
		}
		if (j < 0) (ng -> sub[i]) -> max = 0;
		else {
			(ng -> sub[i]) -> max = 1;
			addtree(ng -> sub[i], (char *)tmp, -1, '\0');
			(ng -> max)--;
			for (j = i; j < ng -> max; j++)
				ng -> sub[j] = ng -> sub[j + 1];
		}
	}
	return(i);
}

static VOID sortng(ng)
ngtree *ng;
{
	long i, j, s;

	s = startng(ng);
	qsort(&(ng -> sub[s]), ng -> max - s, sizeof(ngtree *), cmptree);
	for (i = ng -> max - 1; i > s; i--) {
		if (!ngcmp(ng -> sub[i], ng -> sub[i - 1])) {
			concatng(ng -> sub[i - 1], ng -> sub[i]);
			(ng -> max)--;
			for (j = i; j < ng -> max; j++)
				ng -> sub[j] = ng -> sub[j + 1];
		}
	}
}

static VOID r_sortng(ng)
ngtree *ng;
{
	long i;

	sortng(ng);
	for (i = startng(ng); i < ng -> max; i++)
		if ((ng -> sub[i]) -> max) r_sortng(ng -> sub[i]);
}

static VOID mergeng(ng)
ngtree *ng;
{
	long i, j, k, s;

	s = startng(ng);
	for (i = ng -> max - 1; i > s; i--) {
		for (j = i - 1; j >= s; j--)
		if (!ngcmp(ng -> sub[i], ng -> sub[j])) {
			combineng(ng -> sub[j], ng -> sub[i]);
			(ng -> max)--;
			for (k = i; k < ng -> max; k++)
				ng -> sub[k] = ng -> sub[k + 1];
			break;
		}
	}
}

static VOID r_mergeng(ng)
ngtree *ng;
{
	long i;

	mergeng(ng);
	for (i = startng(ng); i < ng -> max; i++)
		if ((ng -> sub[i]) -> max) r_mergeng(ng -> sub[i]);
}

static VOID r_setflagng(ng, c)
ngtree *ng;
int c;
{
	long i;

	if (!(ng -> max)) {
		ngflag(ng) = c;
		return;
	}
	for (i = startng(ng); i < ng -> max; i++) r_setflagng(ng -> sub[i], c);
}

static int readrc(VOID_A)
{
	FILE *fp;
	char *line;
	long ofs, opt;

	if (!(fp = fopen(newsrcfile, "r"))) return(-1);
	ofs = ftell(fp);

	if (!(topng = (ngtree *)malloc(sizeof(ngtree)))) error(ER_NOMEM);
	if (!(options = (long *)malloc(BUFUNIT * sizeof(long))))
		error(ER_NOMEM);
	opt = 0;

	topng -> name = NULL;
	topng -> offset = -1;
	topng -> max = 0;
	topng -> sub = NULL;

	while ((line = fgetline(fp))) {
		if (*line == ' ' || *line == '\t');
		else if (strncmp(line, IDOPTIONS, sizeof(IDOPTIONS) - 1)) {
			if (addng(line, ofs) < 0) {
				free(line);
				break;
			}
		}
		else {
			if ((opt + 1) % BUFUNIT) {
				options = (long *)realloc(options,
					b_size(opt + 1, long));
				if (!options) error(ER_NOMEM);
			}
			options[opt++] = ofs;
		}
		ofs = ftell(fp);
		free(line);
	}
	fclose(fp);
	if (line || !(topng -> max)) {
		freetree(topng);
		if (options) free(options);
		return(-1);
	}
	options[opt] = -1;
	return(0);
}

static int _dumpng(ng, fpin, fpout)
ngtree *ng;
FILE *fpin, *fpout;
{
	long i, offset;
	int c;

	for (i = startng(ng); i < ng -> max; i++) {
		if ((ng -> sub[i]) -> max) {
			if ((c = _dumpng(ng -> sub[i], fpin, fpout)) < 0)
				return(c);
		}
		else {
			if (*ngname(ng -> sub[i]) == '.')
				offset = ng -> offset;
			else offset = (ng -> sub[i]) -> offset;
			c = fcopyline(offset, ngflag(ng -> sub[i]),
				fpin, fpout);
			if (c < 0) return(c);
		}
	}
	return(0);
}

static VOID dumpng(VOID_A)
{
	static char *orgfile = NULL;
	FILE *fpin, *fpout;
	long i;
	int c;

	if (!orgfile) orgfile = newsrcfile;

	if (!(fpin = fopen(orgfile, "r"))) error(ER_NOREAD);
	if (!(fpout = fopen(temporaryfile, "w"))) error(ER_NOWRITE);

	c = 0;
	if (options) for (i = 0; options[i] >= 0; i++)
		if ((c = fcopyline(options[i], '\0', fpin, fpout)) < 0) break;
	if (!c && topng) c = _dumpng(topng, fpin, fpout);
	fclose(fpin);
	fclose(fpout);
	if (c < 0) {
		unlink(temporaryfile);
		error((i == -1) ? ER_NOREAD : ER_NOWRITE);
	}

	if (orgfile == newsrcfile) {
		if (rename(newsrcfile, backupfile) < 0) error(ER_NOREAD);
		orgfile = backupfile;
	}
	if (rename(temporaryfile, newsrcfile) < 0) error(ER_NOWRITE);
}

static VOID title(VOID_A)
{
	char *cp, *eol;

	locate(0, 0);
	putterm(t_standout);
	cputs2(TITLESTR);
	if (!(cp = strchr(version, ' '))) cp = version;
	while (*(++cp) == ' ');
	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
	cprintf2("%-*.*s", (int)(eol - cp), (int)(eol - cp), cp);
	cprintf2("%-*.*s", n_column - (sizeof(TITLESTR) - 1) - (int)(eol - cp),
		n_column - (sizeof(TITLESTR) - 1) - (int)(eol - cp),
		COPYRIGHTSTR);
	putterm(end_standout);
	tflush();
}

static int listupgroup(ng, start)
ngtree *ng;
long start;
{
	ngtree *sub;
	long i, j, k;
	int c, y;

	locate(0, 1);
	putterm(l_clear);
	cputs2("Newsgroup: ");
	catpath(ng);

	for (y = 0; y < WBODY; y++) {
		locate(0, y + WHEADER);
		putterm(l_clear);
	}

	y = 0;
	for (i = start; i < ng -> max; i++, y++) {
		if (y >= WBODY) break;
		if (i < startng(ng)) {
			locate(3, y + WHEADER);
			cputs2("<..>");
#if	DEBUG
		sub = ng -> sub[0];
		cprintf2(",%p,%p,%p", sub, sub -> name, sub -> sub);
#endif
			continue;
		}

		locate(4, y + WHEADER);
		sub = ng -> sub[i];
		if (strlen(ngname(sub)) <= NAMEFIELD) cputs2(ngname(sub));
		else {
			c = ngname(sub)[NAMEFIELD - 3];
			ngname(sub)[NAMEFIELD - 3] = '\0';
			cputs2(ngname(sub));
			cputs2("...");
			ngname(sub)[NAMEFIELD - 3] = c;
		}
		k = 0;
		for (j = startng(ng); j < i; j++)
			if (!ngcmp(sub, ng -> sub[j])) k++;
		if (k) cprintf2("(%d)", ++k);
#if	DEBUG
		cprintf2(",%p,%p,%p", sub, sub -> name, sub -> sub);
#endif
		if (sub -> max) {
			putch2('>');
			locate(3, y + WHEADER);
			putch2('<');
		} else if (*(sub -> name) == '!') {
			locate(0, y + WHEADER);
			putch2('U');
		}
	}
	return(y);
}

static VOID listupstack(stack, sp)
ngtree **stack;
long sp;
{
	long i;

	for (i = 0; i < sp; i++) {
		locate(NAMEFIELD + 6, i + WHEADER);
		putterm(t_standout);
		catpath(stack[i]);
		if (stack[i] -> max) putch2('.');
		putterm(end_standout);
	}
}

static int helpline(y, key, mes)
int y;
char *key, *mes;
{
	locate(4, y + WHEADER);
	cputs2(key);
	locate(16, y + WHEADER);
	cputs2(mes);
	if (++y >= WBODY) {
		locate(17, n_line - WFOOTER);
		tflush();
		getkey2(0);
		for (y = 0; y < WBODY; y++) {
			locate(0, y + WHEADER);
			putterm(l_clear);
		}
		y = 0;
	}
	return(y);
}

static VOID helpwin(VOID_A)
{
	int y;

	for (y = 0; y < WBODY; y++) {
		locate(0, y + WHEADER);
		putterm(l_clear);
	}

	y = 0;
	y = helpline(y, "Up,    ^P", "move cursor upward");
	y = helpline(y, "Down,  ^N", "move cursor downward");
	y = helpline(y, "Right, ^F", "move cursor to the next page");
	y = helpline(y, "Left,  ^B", "move cursor to the previous page");
	y = helpline(y, "Return", "go in to the sub group");
	y = helpline(y, "Bs", "go out to the parent group");
	y = helpline(y, "^S", "incremental search forward");
	y = helpline(y, "^R", "incremental search backward");
	y = helpline(y, "^L", "redraw screen");
	y = helpline(y, "Space", "toggle subscribed/unsubscribed");
	y = helpline(y, "+", "subscribe recursively");
	y = helpline(y, "-", "unsubscribed recursively");
	y = helpline(y, "s", "sort current category");
	y = helpline(y, "S", "sort current category downward recursively");
	y = helpline(y, "m", "merge current category");
	y = helpline(y, "M", "merge current category downward recursively");
	y = helpline(y, "Del,   ]", "push current group on stack");
	y = helpline(y, "Ins,   [", "pop stack after current group");
	y = helpline(y, "i", "pop stack before current group");
	y = helpline(y, "a", "abandon all stacks");
	y = helpline(y, "d", "delete recursively");
	y = helpline(y, "r", "read .newsrc file");
	y = helpline(y, "w", "write .newsrc file");
	y = helpline(y, "W", "write .newsrc file, and quit");
	y = helpline(y, "Esc,   q", "quit");
	y = helpline(y, "?,     h", "show this");
	if (y) {
		locate(17, n_line - WFOOTER);
		tflush();
		getkey2(0);
	}
	locate(0, n_line - WFOOTER);
	putterm(l_clear);
}

static long searchmove(ng, no, c)
ngtree *ng;
long no;
int c;
{
	static char *buf = NULL;
	long pos;
	int i, n, s, len;

	pos = no;
	if (isearch > 0) {
		if (pos <= startng(ng)) pos = startng(ng);
		n = isearch;
		s = 1;
	}
	else {
		if (pos <= startng(ng)) pos = ng -> max - 1;
		n = -isearch;
		s = -1;
	}

	if (!buf) {
		if (!(buf = (char *)malloc(BUFUNIT))) error(ER_NOMEM);
		*buf = '\0';
	}
	len = strlen(buf);

	switch (c) {
		case CTRL('S'):
			if (!len) isearch = 0;
			else if (n > 2) pos++;
			else if (n > 1) pos = startng(ng);
			s = 1;
			break;
		case CTRL('R'):
			if (!len) isearch = 0;
			else if (n > 2) pos--;
			else if (n > 1) pos = ng -> max - 1;
			s = -1;
			break;
		case K_BS:
			if (!len) isearch = 0;
			else buf[--len] = '\0';
			break;
		default:
			if (n == 1) buf[len = 0] = '\0';
			if (c < ' ' || c == '\177' || c >= K_MIN) isearch = 0;
			else {
				buf[len++] = c;
				if (!((len + 1) % BUFUNIT)) {
					buf = (char *)realloc(buf,
						b_size(len + 1, char));
					if (!buf) error(ER_NOMEM);
				}
				buf[len] = '\0';
			}
			break;
	}

	if (!isearch) {
		if (c == K_CR) return(-1);
		else if (c != K_ESC) putterm(t_bell);
		putterm(end_standout);
		locate(0, (no % WBODY) + WHEADER);
		return(no);
	}

	i = 0;
	if (pos >= startng(ng) && pos < ng -> max) for (;;) {
		if (!strncmp(ngname(ng -> sub[pos]), buf, len)) {
			i = 1;
			break;
		}
		pos += s;
		if (pos < startng(ng) || pos >= ng -> max) break;
	}

	if (!i) cputs2("Failing ");
	cputs2("I-search");
	if (s < 0) cputs2(" backward");
	putch2(':');
	putterm(end_standout);
	cputs2(buf);
	if (i) no = pos;
	else if (n != 2 && c != K_BS) putterm(t_bell);
	isearch = s * (i + 2);
	return(no);
}

static int getyesno(prompt)
char *prompt;
{
	int c;

	cputs2(prompt);
	putterm(end_standout);
	for (;;) {
		tflush();
		c = getkey2(0);
		if (c == K_ESC || c == 'n' || c == 'N') {
			locate(0, n_line - WFOOTER);
			putterm(l_clear);
			tflush();
			return(0);
		}
		else if (c == K_CR || c == 'y' || c == 'Y') {
			locate(0, n_line - WFOOTER);
			putterm(l_clear);
			putterm(t_standout);
			tflush();
			return(1);
		}
		putterm(t_bell);
	}
}

static VOID browsegroup(VOID_A)
{
	ngtree *ng, **stack, *tmp;
	long i, top, otop, sp, *opt;
	int c, y, oy, max;

	ng = topng;
	stack = NULL;
	top = sp = 0;
	y = oy = 0;
	title();
	max = listupgroup(ng, top);

	locate(0, n_line - WFOOTER);
	putterm(l_clear);

	locate(1, y + WHEADER);
	putch2('>');
	locate(1, y + WHEADER);
	tflush();

	for (;;) {
		otop = top;
		tmp = NULL;
		oy = y;
		c = getkey2(0);
		locate(0, n_line - WFOOTER);
		putterm(l_clear);
		putterm(t_standout);
		if (isearch) {
			top = searchmove(ng, top + y, c);
			if (top < 0) top = otop;
			else {
				y = top % WBODY;
				top -= y;
				if (top != otop) tmp = ng;
				otop = -1;
			}
		}
		if (otop >= 0) switch(c) {
			case CTRL('P'):
				c = K_UP;
			case K_UP:
				if (y > 0) {
					y--;
					putterm(end_standout);
					break;
				}
			case CTRL('B'):
			case K_LEFT:
				if (top < WBODY) {
					cputs2("cannot move upward:");
					y = -1;
				}
				else {
					top -= WBODY;
					if (c == K_UP) y = WBODY - 1;
					tmp = ng;
				}
				putterm(end_standout);
				break;
			case CTRL('N'):
				c = K_DOWN;
			case K_DOWN:
				if (y < max - 1) {
					y++;
					putterm(end_standout);
					break;
				}
			case CTRL('F'):
			case K_RIGHT:
				if (top + max >= ng -> max) {
					cputs2("cannot move downward:");
					y = -1;
				}
				else {
					top += max;
					if (c == K_DOWN) y = 0;
					tmp = ng;
				}
				putterm(end_standout);
				break;
			case K_BS:
				if (!(ng -> name)) {
					cputs2("cannot go out:");
					putterm(end_standout);
					y = -1;
				}
				else {
					tmp = ng;
					ng = ng -> sub[0];
					cputs2("come out:");
					putterm(end_standout);
					catpath(ng);
				}
				break;
			case CTRL('L'):
				cputs2("redrawn:");
				putterm(end_standout);
				catpath(ng);
				title();
				tmp = ng;
				break;
			case K_CR:
				if (top + y < startng(ng)) {
					tmp = ng;
					ng = ng -> sub[0];
					cputs2("come out:");
					putterm(end_standout);
					catpath(ng);
				}
				else if ((ng -> sub[top + y]) -> max <= 0) {
					cputs2("cannot go in:");
					putterm(end_standout);
					catpath(ng -> sub[top + y]);
					y = -1;
				}
				else {
					tmp = ng = ng -> sub[top + y];
					y = 0;
					top = 0;
					cputs2("come in:");
					putterm(end_standout);
					catpath(ng);
				}
				break;
			case CTRL('R'):
				isearch = -1;
				cputs2("I-search backward:");
				putterm(end_standout);
				break;
			case CTRL('S'):
				isearch = 1;
				cputs2("I-search:");
				putterm(end_standout);
				break;
			case K_DC:
			case ']':
			case '\177':
				if (top + y < startng(ng)) {
					cputs2("cannot push:");
					putterm(end_standout);
					cputs2("..");
					y = -1;
				}
				else if (!(ng -> max)) {
					cputs2("cannot push:");
					putterm(end_standout);
					cputs2("(none)");
					y = -1;
				}
				else if (sp >= WBODY) {
					cputs2("overflow:");
					putterm(end_standout);
					cputs2("all stack");
					y = -1;
				}
				else {
					tmp = ng;
					stack = pushng(ng, top + y, stack, sp);
					cputs2("pushed:");
					putterm(end_standout);
					catpath(stack[sp++]);
				}
				break;
			case K_IC:
			case '[':
				if (sp <= 0) {
					cputs2("cannot pop:");
					putterm(end_standout);
					cputs2("(none)");
					y = -1;
				}
				else {
					cputs2("poped:");
					putterm(end_standout);
					catpath(stack[--sp]);
					i = top + y + 1;
					if (i > ng -> max) i = ng -> max;
					tmp = ng;
					ng = popng(tmp, i, stack[sp]);
				}
				break;
			case ' ':
				if (top + y < startng(ng)) {
					cputs2("cannot subscribe:");
					putterm(end_standout);
					cputs2("..");
					y = -1;
				}
				else if ((ng -> sub[top + y]) -> max > 0) {
					cputs2("cannot subscribe:");
					putterm(end_standout);
					catpath(ng -> sub[top + y]);
					y = -1;
				}
				else if (ngflag(ng -> sub[top + y]) == '!') {
					cputs2("subscribed:");
					putterm(end_standout);
					catpath(ng -> sub[top + y]);
					ngflag(ng -> sub[top + y]) = ':';
					locate(0, y + WHEADER);
					putch2(' ');
					oy = -1;
				}
				else {
					cputs2("unsubscribed:");
					putterm(end_standout);
					catpath(ng -> sub[top + y]);
					ngflag(ng -> sub[top + y]) = '!';
					locate(0, y + WHEADER);
					putch2('U');
					oy = -1;
				}
				break;
			case '+':
				if (top + y < startng(ng)) {
					cputs2("cannot subscribe:");
					putterm(end_standout);
					cputs2("..");
					y = -1;
				}
				else if (!getyesno("subscribe all? [y/n]:"))
					oy = -1;
				else {
					r_setflagng(ng -> sub[top + y], ':');
					cputs2("subscribed:");
					putterm(end_standout);
					catpath(ng -> sub[top + y]);
					if ((ng -> sub[top + y]) -> max > 0)
						oy = -1;
					else tmp = ng;
				}
				break;
			case '-':
				if (top + y < startng(ng)) {
					cputs2("cannot unsubscribe:");
					putterm(end_standout);
					cputs2("..");
					y = -1;
				}
				else if (!getyesno("unsubscribe all? [y/n]:"))
					oy = -1;
				else {
					r_setflagng(ng -> sub[top + y], '!');
					cputs2("unsubscribed:");
					putterm(end_standout);
					catpath(ng -> sub[top + y]);
					if ((ng -> sub[top + y]) -> max > 0)
						oy = -1;
					else tmp = ng;
				}
				break;
			case 'd':
				if (top + y < startng(ng)) {
					cputs2("cannot delete:");
					putterm(end_standout);
					cputs2("..");
					y = -1;
				}
				else if (!getyesno("delete? [y/n]:")) oy = -1;
				else {
					cputs2("deleted:");
					putterm(end_standout);
					tmp = ng -> sub[top + y];
					deleteng(ng, top + y);
					catpath(tmp);
					freetree(tmp);
					tmp = ng;
				}
				break;
			case 'a':
			case 'A':
				if (sp <= 0) {
					cputs2("cannot abandon stack:");
					putterm(end_standout);
					cputs2("(none)");
					y = -1;
				}
				else {
					restorestack(stack, sp);
					free(stack);
					stack = NULL;
					sp = 0;
					tmp = ng;
					cputs2("stack abandoned:");
					putterm(end_standout);
					cputs2("all stack");
				}
				break;
			case '?':
			case 'h':
				cputs2("help:");
				putterm(end_standout);
				cputs2("push any key");
				helpwin();
				tmp = ng;
				break;
			case 'i':
			case 'I':
				if (sp <= 0) {
					cputs2("no stack:");
					putterm(end_standout);
					y = -1;
				}
				else if (top + y < startng(ng)) {
					cputs2("cannot pop:");
					putterm(end_standout);
					cputs2("..");
					y = -1;
				}
				else {
					cputs2("poped:");
					putterm(end_standout);
					catpath(stack[--sp]);
					tmp = ng;
					ng = popng(tmp, top + y, stack[sp]);
					if (++y >= WBODY) {
						top += WBODY;
						y = 0;
					}
				}
				break;
			case 'm':
				mergeng(ng);
				cputs2("merged:");
				putterm(end_standout);
				catpath(ng);
				tmp = ng;
				break;
			case 'M':
				r_mergeng(ng);
				cputs2("merged recursively:");
				putterm(end_standout);
				catpath(ng);
				tmp = ng;
				break;
			case 'q':
			case 'Q':
			case K_ESC:
				if (!getyesno("quit? [y/n]:")) oy = -1;
				else {
					cputs2("quit:\r\n");
					c = -1;
				}
				break;
			case 'r':
				if (stack) {
					restorestack(stack, sp);
					free(stack);
					stack = NULL;
					sp = 0;
				}
				else if (!getyesno("read? [y/n]:")) oy = -1;
				else {
					opt = options;
					tmp = topng;
					if (readrc() < 0) {
						topng = tmp;
						options = opt;
						tmp = ng;
						cputs2("cannot read:");
						putterm(end_standout);
						cputs2(newsrcfile);
						y = -1;
					}
					else {
						freetree(tmp);
						if (opt) free(opt);
						top = 0;
						y = 0;
						tmp = ng = topng;
						cputs2("read:");
						putterm(end_standout);
						cputs2(newsrcfile);
					}
				}
				break;
			case 's':
				sortng(ng);
				cputs2("sorted:");
				putterm(end_standout);
				catpath(ng);
				tmp = ng;
				break;
			case 'S':
				r_sortng(ng);
				cputs2("sorted recursively:");
				putterm(end_standout);
				catpath(ng);
				tmp = ng;
				break;
			case 'w':
			case 'W':
				if (sp > 0) {
					cputs2("not empty:");
					putterm(end_standout);
					cputs2("all stack");
					y = -1;
				}
				else if (!getyesno("write? [y/n]:")) oy = -1;
				else {
					dumpng();
					cputs2("written:");
					putterm(end_standout);
					cputs2(newsrcfile);
					oy = -1;
					if (c == 'W') c = -1;
				}
				break;
			default:
				cputs2("no command binded:");
				putterm(end_standout);
				if (c < ' ') cprintf2("ctrl-%c", c + '@');
				else if (c < '\177') putch2(c);
				else cputs2("(?)");
				y = -1;
				break;
		}
		if (c < 0) break;
		if (y < 0) {
			putterm(t_bell);
			locate(1, (y = oy) + WHEADER);
		}
		if (tmp) {
			if (tmp != ng) {
				top = backng(tmp, ng, stack, sp);
				y = top % WBODY;
				top -= y;
			}
			max = listupgroup(ng, top);
			listupstack(stack, sp);
			if (y >= max && max) y = max - 1;
			oy = -1;
		}
		if (y != oy) {
			locate(1, y + WHEADER);
			putch2('>');
			if (oy >= 0) {
				locate(1, oy + WHEADER);
				putch2(' ');
			}
			locate(1, y + WHEADER);
		}
		tflush();
	}

	tflush();
	if (stack) {
		restorestack(stack, sp);
		free(stack);
	}
}

static int evaloption(argc, argv)
int argc;
char *argv[];
{
	char *cp;
	int i;

	for (i = 1; i < argc; i++) {
		cp = argv[i];
		if (*(cp++) != '-') return(i);
		if (*cp == '-') return(i + 1);
		for (; *cp; cp++) switch (*cp) {
			case 'i':
				flags |= FL_INTERACT;
				break;
			case 'm':
				flags |= FL_MERGE;
				break;
			case 'u':
				flags |= FL_NOBACK;
				break;
			case '?':
			case 'h':
				usage(argv[0]);
				break;
			default:
				fputc('\007', stderr);
				fputs("illegal option with -", stderr);
				fputc(*cp, stderr);
				fputc('\n', stderr);
				usage(argv[0]);
				break;
		}
	}
	return(i);
}

int main(argc, argv)
int argc;
char *argv[];
{
	int i;

#ifdef	DEBUG
	mtrace();
#endif
	i = evaloption(argc, argv);
	if (i + 1 < argc) {
		fputc('\007', stderr);
		usage(argv[0]);
	}
	else if (i < argc) newsrcfile = argv[i];
	else {
		if (cdhome() < 0) {
			fprintf(stderr,
				"%s: cannot change your home directory.\n",
				argv[0]);
			exit(1);
		}
		newsrcfile = NEWSRC;
	}
	if (readrc() < 0) error(ER_NOREAD);

	backupfile = (char *)malloc(strlen(newsrcfile) + sizeof(BAKFILE));
	if (!backupfile) error(ER_NOMEM);
	strcpy(backupfile, newsrcfile);
	strcat(backupfile, BAKFILE);

	temporaryfile = (char *)malloc(strlen(newsrcfile) + 10);
	if (!temporaryfile) error(ER_NOMEM);
	sprintf(temporaryfile, "%s.%d", newsrcfile, getpid());

	if (!(flags & FL_INTERACT)) {
		if (flags & FL_MERGE) r_mergeng(topng);
		else r_sortng(topng);
		dumpng();
	}
	else {
		if (flags & FL_MERGE) r_mergeng(topng);
		inittty(0);
		getterment();
		ttyiomode();
		initterm();
		getwsize(80, 3);

		putterms(t_clear);
		browsegroup();
	}

	freetree(topng);
	if (options) free(options);
	if (flags & FL_NOBACK) unlink(backupfile);
	free(backupfile);
	free(temporaryfile);

	exit2(0);
	return(0);
}
